local physics = require ( "physics" )

physics.start ()

local dbg1 = true --transparent face and box id text
local printRuntime = false
--physics.setDrawMode ( "hybrid" )

local allBoxes = {}

--listener to manage touch on boxes
--creates touch joint to move box around
--removes touch joint on event end ( or cancel )
local function touchBox ( event )
	local boxGroup = event.target
	
	if ( event.phase == "began" ) then
		-- Set touch focus
		display.getCurrentStage():setFocus( boxGroup )
		
		boxGroup.isFixedRotation = true
		boxGroup.rotation = 0
		
		boxGroup.touchJoint = physics.newJoint( "touch", boxGroup, boxGroup.x, boxGroup.y )
		boxGroup.touchJoint.maxForce = 5000
		
	elseif ( event.phase == "moved" ) then
		boxGroup.touchJoint:setTarget ( event.x, event.y )
		
	elseif ( event.phase == "ended" or event.phase == "cancelled" ) then
		display.remove ( boxGroup.touchJoint )
		boxGroup.isFixedRotation = false
		-- Reset touch focus
		display.getCurrentStage():setFocus( nil )

	end
	return true
end

--===============DRAW FUNCTIONS called from runtime enterFrame listener

--takes table in appropriate format and applies its values to corners or images
--requires allBoxes[i] of box
--tab argument is optional: if not provided, angle 0 is set
local function rearangeCorners ( box, tab )
	local CornTab = tab or box.corners[1]
	for img, valuesTable in pairs ( CornTab ) do
		for prop, value in pairs ( valuesTable ) do
			box[img]["path"][prop] = value
		end
	end
end
	
local function printOrder ( parameter )
	print ( "ON KEY PRESS" )
	local boxesToReorder = {}
	--make list of all boxes that should be reordered
	for i = 1, #allBoxes do
		if true then --allBoxes[i].group.isAwake then
			boxesToReorder[#boxesToReorder+1] = allBoxes[i]
		end
	end
	
	--add property order to every box to reorder. This property rises with boxes to right and above this box.
	for i = 1, #boxesToReorder do
		local thisBox = boxesToReorder[i]
		--every box is checked multiple times so we calculate values only once and store them as property of face.
		--we need to remove this property later
		if not thisBox.face.L then thisBox.face.L = thisBox.group.x end-- - thisBox.face.width/2 end --left side of face
		if not thisBox.face.R then thisBox.face.R = thisBox.group.x end-- + thisBox.face.width/2 end --ride side of face
		if not thisBox.face.U then thisBox.face.U = thisBox.group.y end-- - thisBox.face.height/2 end --up side of face
		if not thisBox.face.D then thisBox.face.D = thisBox.group.y end-- + thisBox.face.height/2 end --down side of face
		--setup property to store draw order
		if not thisBox.order then thisBox.order = 1 end
		
		if not parameter then print ( i, thisBox.id ) end
		if parameter == i then print ( i, thisBox.face.L, thisBox.face.R, thisBox.face.U, thisBox.face.D ) end
		
		--if i == 1 then print ( i..":") end
		for a = 1, #boxesToReorder do
			local otherBox = boxesToReorder[a]
			if not ( thisBox == otherBox ) then
				if not otherBox.face.L then otherBox.face.L = otherBox.group.x - otherBox.face.width/2 end --left side of face
				if not otherBox.face.R then otherBox.face.R = otherBox.group.x + otherBox.face.width/2 end --ride side of face
				if not otherBox.face.U then otherBox.face.U = otherBox.group.y - otherBox.face.height/2 end --up side of face
				if not otherBox.face.D then otherBox.face.D = otherBox.group.y + otherBox.face.height/2 end --down side of face

		--print ( otherBox.face.L, otherBox.face.R, otherBox.face.U, otherBox.face.D )		
				if thisBox.face.R + 1 > otherBox.face.L then
				--	if i == 1 then print ( "   "..a, "R <= L" ) end
					thisBox.order = thisBox.order + 1
					if not parameter then print ( "   is right from " .. a) end
					if not parameter then print ( "        "..(thisBox.face.R + 1).." "..otherBox.face.L) end
				end
				if thisBox.face.D - 1 < otherBox.face.U then
				--	if i == 1 then print ( "   "..a, "D <= U" ) end
					thisBox.order = thisBox.order + 1
					if not parameter then print ( "   is above from " .. a ) end
					if not parameter then print ( "        "..(thisBox.face.D - 1).." "..otherBox.face.U) end
				end
			end
		end
	end
	
	--remove properties added in previous for loop
	for i = 1, #boxesToReorder do
		local thisBox = boxesToReorder[i]
		--print ( "result order: "..table.indexOf( allBoxes, thisBox ), thisBox.order )
		if not parameter then print ( thisBox.id, thisBox.order ) end
		thisBox.face.L, thisBox.face.R, thisBox.face.U, thisBox.face.D, thisBox.order = nil, nil, nil, nil, nil
	end
	if not parameter then print ( "=====" ) end
	
end

--rearanges boxes so they are positioned right ( front or back )
--this could use some optimalization or completely different approach
local function rearangeDrawOrder ()
	local boxesToReorder = {}
	--make list of all boxes that should be reordered
	for i = 1, #allBoxes do
		if true then --allBoxes[i].group.isAwake then
			boxesToReorder[#boxesToReorder+1] = allBoxes[i]
		end
	end
	
	--add property order to every box to reorder. This property rises with boxes to right and above this box.
	for i = 1, #boxesToReorder do
		local thisBox = boxesToReorder[i]
		--every box is checked multiple times so we calculate values only once and store them as property of face.
		--we need to remove this property later
		if not thisBox.face.L then thisBox.face.L = thisBox.group.x end --left side of face
		if not thisBox.face.R then thisBox.face.R = thisBox.group.x end --ride side of face
		if not thisBox.face.U then thisBox.face.U = thisBox.group.y end --up side of face
		if not thisBox.face.D then thisBox.face.D = thisBox.group.y end --down side of face
		--setup property to store draw order
		if not thisBox.order then thisBox.order = 1 end
		
		if printRuntime then print ( thisBox.id..":") end
		
		for a = 1, #boxesToReorder do
			local otherBox = boxesToReorder[a]
			if not ( thisBox == otherBox ) then
				if not otherBox.face.L then otherBox.face.L = otherBox.group.x end --left side of face
				if not otherBox.face.R then otherBox.face.R = otherBox.group.x end --ride side of face
				if not otherBox.face.U then otherBox.face.U = otherBox.group.y end --up side of face
				if not otherBox.face.D then otherBox.face.D = otherBox.group.y end --down side of face
		--print ( thisBox.face.L, thisBox.face.R, thisBox.face.U, thisBox.face.D )
		--print ( otherBox.face.L, otherBox.face.R, otherBox.face.U, otherBox.face.D )		
				if thisBox.face.R + 1 > otherBox.face.L then
					thisBox.order = thisBox.order + 1
					if printRuntime then print ( "   is right from " .. a ) end
				end
				if thisBox.face.D - 1 < otherBox.face.U then
					thisBox.order = thisBox.order + 1
					if printRuntime then print ( "   is above from " .. a ) end
				end
			end
		end
	end

	--remove properties added in previous for loop
	for i = 1, #boxesToReorder do
		local thisBox = boxesToReorder[i]
		--print ( "result order: "..thisBox.id, thisBox.order )
		
		thisBox.face.L, thisBox.face.R, thisBox.face.U, thisBox.face.D = nil, nil, nil, nil
	end
	--print ( "=====================================")
	if printRuntime then print ( "---RENDER ORDER") end
	--find box with lowest order value and bring it to front ( then bring front other boxes )
	while #boxesToReorder > 0 do
		local resBox = boxesToReorder[#boxesToReorder]
		local resBoxIndex = #boxesToReorder
		--find lowest order value
		for i = 1, #boxesToReorder do
			if not ( boxesToReorder[i] == resBox ) then
				if resBox.order > boxesToReorder[i].order then resBox = boxesToReorder[i] resBoxIndex = i end
			end
		end
		--print ( table.indexOf( allBoxes, resBox ) , resBox.order )
		if printRuntime then print ( resBox.id, resBox.order ) end
		resBox.group:toFront()
		resBox.order = nil
		table.remove ( boxesToReorder, resBoxIndex )
	end
	if printRuntime then print ( "=====================================") end
end

--==============CREATE OBJECTS

--background rect
local background = display.newRect ( display.contentCenterX + 30, 500-245+255/2, 1000, 255 )


--create squares that act as limits
limitsBoxes = {}
--down horizontal
limitsBoxes[1] = display.newRect( 510, 500, 1000, 20 )
limitsBoxes[1].alpha = 0.01
--up horizontal
limitsBoxes[2] = display.newRect( 510, 245, 1000, 20 )
limitsBoxes[2].alpha = 0.01
--right vertical
limitsBoxes[3] = display.newRect( 1000, 370, 20, 220 )
limitsBoxes[3].alpha = 0.01
--left vertical
limitsBoxes[4] = display.newRect( 20, 370, 20, 220 )
limitsBoxes[4].alpha = 0.01

--add limit to physics
for i = 1, #limitsBoxes do
	physics.addBody( limitsBoxes[i], "static", { friction = 1 } )
end

--create boxes
for i = 1, 4 do
--[[
We will create group for each box and put all parts of box into this group.
We will add this group later as physics body.
Because of this, all parts of group move together and we have everything in local coordinates.
]]
	local mainName = "box_big"
	
	allBoxes[i] = {}
	allBoxes[i].id = i
	allBoxes[i].group = display.newGroup()
	allBoxes[i].group.x = 200 + 200 * ( i - 1)
	allBoxes[i].group.y = 380
	
	
	allBoxes[i].face = display.newImageRect( allBoxes[i].group, mainName.."_face.png", 60,58)
	if dbg1 then
		allBoxes[i].face.alpha = 0.35
		allBoxes[i].text = display.newText( { parent = allBoxes[i].group, text = i, x = 0, y = 0, fontSize = 20	} )
		allBoxes[i].text:setFillColor ( 0 )
	end
	
	allBoxes[i].left = display.newImageRect( allBoxes[i].group, mainName.."_left.png", 16,58)
	allBoxes[i].left.x = allBoxes[i].face.x - allBoxes[i].face.width/2
	allBoxes[i].left.anchorX = 1
	
	
	allBoxes[i].right = display.newImageRect( allBoxes[i].group, mainName.."_right.png", 16,58)
	allBoxes[i].right.x = allBoxes[i].face.x + allBoxes[i].face.width/2
	allBoxes[i].right.anchorX = 0

	allBoxes[i].top = display.newImageRect( allBoxes[i].group, mainName.."_top.png", 60,12)
	allBoxes[i].top.y = allBoxes[i].face.y - allBoxes[i].face.height/2
	allBoxes[i].top.anchorY = 1
	
	allBoxes[i].bottom = display.newImageRect( allBoxes[i].group, mainName.."_bottom.png", 60,12)
	allBoxes[i].bottom.y = allBoxes[i].face.y + allBoxes[i].face.height/2
	allBoxes[i].bottom.anchorY = 0

	--We need 4 key values for box at 0, 90, 180 and 270 degrees rotation.
	--Values between key values will be calculated on run by reDrawBoxes on enterFrame event.

		local LW = allBoxes[i].left.width
		local LH = allBoxes[i].left.height
		local RW = allBoxes[i].right.width
		local RH = allBoxes[i].right.height
		
		local TH = allBoxes[i].top.height
		local BH = allBoxes[i].bottom.height
		
	allBoxes[i].corners = {}
	
	--	at 0 degrees: initial value, siting on bottom
	allBoxes[i].corners[1] = { 
		left = { x1 = 2*LW, y1 = -TH, x2 = 2*LW, y2 = -BH },
		right = { x4 = 0, y4 = -TH, x3 = 0, y3 = -BH },
		top = { x1 = LW, y1 = 0, x4 = RW, y4 = 0 },
		bottom = { x2 = LW, y2 = -2*BH, x3 = RW, y3 = -2*BH }
	}
		--at 90 degr: laying on right side
	allBoxes[i].corners[2] = { 
		left = { x1 = 0, y1 = -TH, x2 = 0, y2 = -BH },
		right = { x4 = -2*RW, y4 = -TH, x3 = -2*RW, y3 = -BH },
		top = { x1 = -LW, y1 = 0, x4 = -RW, y4 = 0 },
		bottom = { x2 = -LW, y2 = -2*BH, x3 = -RW, y3 = -2*BH },
	}
		--at 180 degr: laying on top side
	allBoxes[i].corners[3] = { 
		left = { x1 = 0, y1 = BH, x2 = 0, y2 = BH },
		right = { x4 = -2*RW, y4 = TH, x3 = -2*RW, y3 = BH },
		top = { x1 = -LW, y1 = 2*TH, x4 = -RW, y4 = 2*TH },
		bottom = { x2 = -RW, y2 = 0, x3 = -RW, y3 = 0 },
	}
		--at 270 degr: laying on left side
		allBoxes[i].corners[4] = { 
		left = { x1 = 2*LW, y1 = TH, x2 = 2*LW, y2 = BH },
		right = { x4 = 0, y4 = TH, x3 = 0, y3 = BH },
		top = { x1 = RW, y1 = 2*TH, x4 = LW, y4 = 2*TH },
		bottom = { x2 = LW, y2 = 0, x3 = RW, y3 = 0 },
	}

end
--=========================PHYSICS BODIES
--Add boxes to physics. We will add whole group containing 5 pictures that put together box.
for i = 1, #allBoxes do
	local thisBox = allBoxes[i]
	
	--We need to redo physics body shape so it only covers face of the box.
	local W, H = thisBox.face.width/2, thisBox.face.height/2
	local bodyDesc = { -W,-H, -W,H, W,-H, W,H }
	
	physics.addBody ( thisBox.group, "dynamic", { bounce = 0.05, density = 3, friction = 1, shape = bodyDesc } )
	
	thisBox.group:addEventListener ( "touch", touchBox )
end

--========================RUNTIME enterFrame LISTENER
local function reDrawBoxes ()
--this is function that recalculates visual of all boxes in real time and applies it using rearangeCorners and rearangeDrawOrder functions.
	for i = 1, #allBoxes do
		local thisBox = allBoxes[i]
		
		if thisBox.group.isAwake then
			local rotation = allBoxes[i].group.rotation
			local angle = rotation % 360 --value that represents the same visual rotation but in interval 0 to 360
			
			
			local startTable, endTable --key values will be stored here
			local interval --we will calculate values for each interval of 90...this will contain angle value recalculated to this interval
			
			if angle < 90 then
				startTable = thisBox.corners[1]
				endTable = thisBox.corners[2]
				interval = angle
				
				thisBox.top:toFront()
				thisBox.bottom:toBack()
			elseif angle >=90 and angle < 180 then
				startTable = thisBox.corners[2]
				endTable = thisBox.corners[3]
				interval = angle - 90
				
				thisBox.left:toFront()
				thisBox.right:toBack()
			elseif angle >= 180 and angle < 270 then
				startTable = thisBox.corners[3]
				endTable = thisBox.corners[4]
				interval = angle - 180
				
				thisBox.bottom:toFront()
				thisBox.top:toBack()
			elseif angle >=270 and angle < 360 then
				startTable = thisBox.corners[4]
				endTable = thisBox.corners[1]
				interval = angle - 270
				
				thisBox.right:toFront()
				thisBox.left:toBack()
			end
			thisBox.face:toFront()
			
			local actualTable = {} --actual calculated values for given rotation
			
			for img, valuesTable in pairs ( startTable ) do
				
				actualTable[img] = {}

				for prop, value in pairs ( valuesTable ) do 

					local startTableValue = startTable[img][prop] --value in startTable table
					local endTableValue = endTable[img][prop]
					
					local actualTableValue = ( endTableValue - startTableValue )*( interval / 90 ) + startTableValue
					
					--put calculated actualTableValue to table
					actualTable[img][prop] = actualTableValue
				end
			end
			--apply calculated values to thisBox corners
			rearangeCorners ( thisBox, actualTable )
		end
	end --end of loop for allBoxes
	
	--redo order in which are objects on screen. We need faces to be always in front of sides.
	rearangeDrawOrder ()
end

Runtime:addEventListener( "enterFrame", reDrawBoxes )

local function keyPress ( event )
	if event.phase == "up" then
		if event.keyName == "space" then
			printOrder ()
		elseif event.keyName == "1" then
			printOrder ( 1 )
		elseif event.keyName == "2" then
			printOrder ( 2 )
		elseif event.keyName == "3" then
			printOrder ( 3 )
		elseif event.keyName == "4" then
			printOrder ( 4 )
		elseif event.keyName == "r" then
			printRuntime = not printRuntime
		end
	end
	return true
end

Runtime:addEventListener ( "key", keyPress )