local physics = require ( "physics" )

physics.start ()

local P = true

local dbg1 = true --transparent face and box id text
local dbgLines = false --lines on boxes bounderies
local dbgLines2 = true --lines on boxes outer bounderies (group)

--prints on every frame relative position and render order of boxes. 
--Can be run and stopped by keys ( "c" and "space")
local printRuntime = false


--physics.setDrawMode ( "hybrid" )

local allBoxes = {}

--listener to manage touch on boxes
--creates touch joint to move box around
--removes touch joint on event end ( or cancel )
local function touchBox ( event )
	local boxGroup = event.target
	
	if ( event.phase == "began" ) then
		-- Set touch focus
		display.getCurrentStage():setFocus( boxGroup )
		
		-- boxGroup.isFixedRotation = true
		-- boxGroup.rotation = 0
		
		-- boxGroup.touchJoint = physics.newJoint( "touch", boxGroup, boxGroup.x, boxGroup.y )
		-- boxGroup.touchJoint.maxForce = 5000
		
		boxGroup.offsetX = boxGroup.x - event.x
		boxGroup.offsetY = boxGroup.y - event.y
		
	elseif ( event.phase == "moved" ) then
		--if boxGroup.touchJoint then boxGroup.touchJoint:setTarget ( event.x, event.y ) end
		boxGroup.x = event.x + boxGroup.offsetX
		boxGroup.y = event.y + boxGroup.offsetY
	elseif ( event.phase == "ended" or event.phase == "cancelled" ) then
		-- display.remove ( boxGroup.touchJoint )
		-- boxGroup.touchJoint = nil
		-- boxGroup.isFixedRotation = false
		-- Reset touch focus
		display.getCurrentStage():setFocus( nil )

	end
	return true
end

--===============DRAW FUNCTIONS called from runtime enterFrame listener

--takes table in appropriate format and applies its values to corners or images
--requires allBoxes[i] of box
--tab argument is optional: if not provided, angle 0 is set
local function rearangeCorners ( box, tab )
	local CornTab = tab or box.corners[1]
	for img, valuesTable in pairs ( CornTab ) do
		for prop, value in pairs ( valuesTable ) do
			box[img]["path"][prop] = value
		end
	end
end

--rearanges boxes so they are positioned right ( front or back )
--this could use some optimalization or completely different approach
local function rearangeDrawOrder ()
	if dbgLines then
	--debug: show lines on contentBounds
		for i = 1, #allBoxes do
			local thisBox = allBoxes[i]

			if not thisBox.lines then thisBox.lines = {} end
			if i == 1 then
				for lineI = 1, #thisBox.lines do
					display.remove (thisBox.lines[lineI])
				end
				thisBox.lines = nil
				thisBox.lines = {}

				thisBox.lines[#thisBox.lines+1] = display.newLine ( thisBox.face.contentBounds.xMax, 0, thisBox.face.contentBounds.xMax, 1000 )
				thisBox.lines[#thisBox.lines]:setStrokeColor ( 1,0,0 )
				
				thisBox.lines[#thisBox.lines+1] = display.newLine ( thisBox.face.contentBounds.xMin, 0, thisBox.face.contentBounds.xMin, 1000 )
				thisBox.lines[#thisBox.lines]:setStrokeColor ( 0,0,1 )

				thisBox.lines[#thisBox.lines+1] = display.newLine ( 0, thisBox.face.contentBounds.yMax, 1000, thisBox.face.contentBounds.yMax)
				thisBox.lines[#thisBox.lines]:setStrokeColor ( 0 )
				
				thisBox.lines[#thisBox.lines+1] = display.newLine ( 0, thisBox.face.contentBounds.yMin, 1000, thisBox.face.contentBounds.yMin )
				thisBox.lines[#thisBox.lines]:setStrokeColor ( 0,1,0 )
			end
		end
	end
	
	-- find out if two boxes overlap each other
	-- if the overlap, send back the one that isnt left top of those boxes them
	local function overlaps ( thisBox, otherBox )
	
	--returns true if two boxes overlaps
		local a = {
			xMin = thisBox.group.contentBounds.xMin,
			xMax = thisBox.group.contentBounds.xMax,
			yMin = thisBox.group.contentBounds.yMin,
			yMax = thisBox.group.contentBounds.yMax,
		}

		local b = {
			xMin = otherBox.group.contentBounds.xMin,
			xMax = otherBox.group.contentBounds.xMax,
			yMin = otherBox.group.contentBounds.yMin,
			yMax = otherBox.group.contentBounds.yMax,
		}
		
		if
			a.xMax < b.xMin or --a left to b
			a.xMin > b.xMax or --a right to b
			a.yMax < b.yMin or --a above b
			a.yMin > b.yMax then --a below b
			
			return false
		else
			return true
		end
	end
	
	local function findUpBox ( thisBox, otherBox )
	--returns box that should be behing other box in this parameter
		
		local function getSecondCorners ( box )
			-- returns X and Y of every corner with second most extreme value
			local secLeftCorner, secRightCorner, secUpCorner, secDownCorner
	--		1--4
	--		|  |
	--		2--3
	
			--we need value of second most left and right corner ( not contentBounds but second corners )
			local cors = {}
			cors[1] = { box.group:localToContent ( -box.face.width/2, -box.face.height/2 ) }
			cors[2] = { box.group:localToContent ( -box.face.width/2, box.face.height/2 ) }
			cors[3] = { box.group:localToContent ( box.face.width/2, box.face.height/2 ) }
			cors[4] = { box.group:localToContent ( box.face.width/2, -box.face.height/2 ) }
			
			secLeftCorner, secRightCorner, secUpCorner, secDownCorner = { 20000,0 }, { -20000,0 }, { 0,20000 }, { 0,-20000 }
			maxLeftCorner, maxRightCorner, maxUpCorner, maxDownCorner = { 20000,0 }, { -20000,0 }, { 0,20000 }, { 0,-20000 }
			
			for i = 1, 4 do
			-- print ( cors[i][1] )
				if cors[i][1] < maxLeftCorner[1] then
					secLeftCorner = maxLeftCorner
					maxLeftCorner = cors[i]
				elseif ( cors[i][1] < secLeftCorner[1]) and ( cors[i][1] >= maxLeftCorner[1] ) then
					secLeftCorner = cors[i]
				end
				if cors[i][1] > maxRightCorner[1] then
					secRightCorner = maxRightCorner
					maxRightCorner= cors[i]
				elseif ( cors[i][1] > secRightCorner[1] ) and ( cors[i][1] <= maxRightCorner[1] ) then
					secRightCorner = cors[i]
				end
				if cors[i][2] < maxUpCorner[2] then
					secUpCorner = maxUpCorner
					maxUpCorner = cors[i]
				elseif ( cors[i][2] < secUpCorner[2] ) and ( cors[i][2] >= maxUpCorner[2] ) then
					secUpCorner = cors[i]
				end
				if cors[i][2] > maxDownCorner[2] then
					secDownCorner = maxDownCorner
					maxDownCorner= cors[i]
				elseif ( cors[i][2] > secDownCorner[2] ) and ( cors[i][2] <= maxDownCorner[2] ) then
					secDownCorner = cors[i]
				end
			end

			return secLeftCorner, secRightCorner, secUpCorner, secDownCorner
		end
		
		--compute second corners if not done in previous iteration.
		--they are indexed tables. table[1] is x value, table[2] is y value
		if not thisBox.secL then thisBox.secL, thisBox.secR, thisBox.secU, thisBox.secD = getSecondCorners ( thisBox ) end
		if not otherBox.secL then otherBox.secL, otherBox.secR, otherBox.secU, otherBox.secD = getSecondCorners ( otherBox ) end

		-- display.remove ( secLcirc )
		-- display.remove ( secRcirc )
		-- secLcirc = display.newCircle ( thisBox.secU[1], thisBox.secU[2] , 10 )
		-- secRcirc = display.newCircle ( thisBox.secD[1], thisBox.secD[2] , 10 )
		-- secRcirc:setFillColor ( 0,0,1 )
		-- secLcirc:setFillColor ( 1,0,0 )
		
		--booleans with result
		local isLeft, isRight, isAbove = false, false, false
		local incline --boolean that stores if box is inclined to right ( true ) or left ( false ). 
		-- This affects what values we use to use to decide positions.
		
		if math.floor(otherBox.secL[2]) == math.floor(otherBox.face.contentBounds.yMin) then
			incline = true
		else
			incline = false
		end
		
		local rule1, rule2, rule3, rule4
		
		if incline == true then
		-- plus solution
			--above
			rule1 = ( thisBox.face.contentBounds.xMax > otherBox.secL[1] )
			rule2 = ( thisBox.face.contentBounds.yMin < otherBox.face.contentBounds.yMax )
			rule3 = ( thisBox.face.contentBounds.xMax < otherBox.secR[1] ) 
			rule4 = ( thisBox.face.contentBounds.xMin > otherBox.secD[2] )
			
		else
		--minus solution
			rule1 = ( thisBox.face.contentBounds.xMax > otherBox.face.contentBounds.xMin )
			rule2 = ( thisBox.face.contentBounds.yMin < otherBox.secD[2] )
			rule3 = ( thisBox.face.contentBounds.xMax < otherBox.secL[1] )
			rule4 = ( thisBox.face.contentBounds.yMin > otherBox.secU[2] )
			
			--[[
			( thisBox.face.contentBounds.yMax <= otherBox.secU[2] ) and
					( thisBox.face.contentBounds.xMin <= otherBox.secR[1] ) and
					( thisBox.face.contentBounds.xMax >= otherBox.face.contentBounds.xMin )
			
			isRight = ( thisBox.face.contentBounds.yMin <= otherBox.secD[2] ) and
					( thisBox.face.contentBounds.xMin >= otherBox.secR[1] ) and
					( thisBox.face.contentBounds.yMax >= otherBox.face.contentBounds.yMin )
				]]
		end
		
		isAbove = rule1 and rule2 and not (rule3 and rule4)
		
		if P then
			print ( thisBox.id.." => "..otherBox.id.." incline is "..(incline and "+" or "-") )
			print ( "rule1: "..tostring (rule1) )
			print ( "rule2: "..tostring (rule2) ) --..": ".. thisBox.face.contentBounds.yMin .. " < " ..otherBox.secD[2] )
			print ( "not rule 3 and 4: "..tostring (not (rule3 and rule4) ) )
			print ( "rule3: "..tostring (rule3) )
			print ( "rule4: "..tostring (rule4) )
			
			-- display.remove ( bflm )
			-- display.remove ( psvz )
			-- bflm = display.newLine ( 0, thisBox.face.contentBounds.yMin, 1000, thisBox.face.contentBounds.yMin )
			-- psvz = display.newCircle( otherBox.secD[1], otherBox.secD[2], 10 )
		end

		if isLeft then thisBox.leftText.text = thisBox.leftText.text.."  "..otherBox.id end
		if isRight then thisBox.rightText.text = thisBox.rightText.text.."  "..otherBox.id end
		if isAbove then thisBox.aboveText.text = thisBox.aboveText.text.."  "..otherBox.id end
		otherBox.inclineText.text = "incline: ".. ( incline == true and "+" or "-" )
		
		if isRight or isAbove then
			return thisBox, otherBox
		else
			return otherBox, thisBox
		end
	end

	--=========================
	--draw order assesment
	
	--restart draw order
	for i = 1, #allBoxes do
		local box = allBoxes[i]
		--reset values
		box.order = 1
		box.secL, box.secR, box.secU, box.secD = nil,nil,nil,nil
	end
	--setup order value: more forward boxes should have higher order value
	for i = 1, #allBoxes do
		local thisBox = allBoxes[i]
		for a = 1, #allBoxes do
			local otherBox = allBoxes[a]
			if not ( thisBox == otherBox ) then
				if overlaps ( thisBox, otherBox ) then
					local upBox, downBox = findUpBox ( thisBox, otherBox )
					upBox.order = downBox.order + upBox.order
				end
			end
		end
	end
			P = false
	--prepaire table off boxes that need to have order rearanged
	local boxesToReorder = {}
	
	for i = 1, #allBoxes do
		if allBoxes[i].order > 1 then
			boxesToReorder[#boxesToReorder + 1] = allBoxes[i]
		end
	end
	
	--bring forward boxes with lowest order property first
	while #boxesToReorder > 0 do
		local res = boxesToReorder[#boxesToReorder]
		local resID = #boxesToReorder
		
		for i = 1, #boxesToReorder-1 do
			if res.order > boxesToReorder[i].order then
				res = boxesToReorder[i]
				resID = i
			end
		end
		res.group:toFront()
		res.orderText.text = "order: "..res.order
		table.remove ( boxesToReorder, resID)
	end
end

--==============CREATE OBJECTS

--background rect
local background = display.newRect ( display.contentCenterX + 30, 500-245+255/2, 1000, 255 )
background.alpha = 0.5

--create squares that act as limits
limitsBoxes = {}
--down horizontal
limitsBoxes[1] = display.newRect( 510, 500, 1000, 20 )
--limitsBoxes[1].alpha = 0.01
--up horizontal
limitsBoxes[2] = display.newRect( 510, 245, 1000, 20 )
--limitsBoxes[2].alpha = 0.01
--right vertical
limitsBoxes[3] = display.newRect( 1000, 370, 20, 220 )
--limitsBoxes[3].alpha = 0.01
--left vertical
limitsBoxes[4] = display.newRect( 20, 370, 20, 220 )
--limitsBoxes[4].alpha = 0.01

--add limit to physics
for i = 1, #limitsBoxes do
	physics.addBody( limitsBoxes[i], "static", { friction = 1 } )
end

--create boxes
for i = 1, 4 do
--[[
We will create group for each box and put all parts of box into this group.
We will add this group later as physics body.
Because of this, all parts of group move together and we have everything in local coordinates.
]]
local mainName, faceWidth, faceHeight, sideWidth, sideHeight

	if i == 4 then
		mainName = "box_big"
		faceWidth, faceHeight = 60/2, 58/2
		sideWidth, sideHeight = 16/2, 12/2
	elseif i == 3 then
		mainName = "box_big"
		faceWidth, faceHeight = 60, 58/2
		sideWidth, sideHeight = 16, 12/2
	else
		mainName = "box_big"
		faceWidth, faceHeight = 60, 58
		sideWidth, sideHeight = 16, 12
	end
	
	allBoxes[i] = {}
	allBoxes[i].id = i
	allBoxes[i].group = display.newGroup()
	allBoxes[i].group.x = 200 + 200 * ( i - 1)
	allBoxes[i].group.y = 380
	
	
	allBoxes[i].face = display.newImageRect( allBoxes[i].group, mainName.."_face.png", faceWidth, faceHeight )
		
	local color
	if i == 1 then color = { 1,0,0 } elseif i == 2 then color = { 0,1,0 } elseif i == 3 then color = { 0,0,1 } elseif i == 4 then color = {1} end
	
	allBoxes[i].face:setStrokeColor ( unpack( color ) )
	allBoxes[i].face.strokeWidth = 1
	allBoxes[i].iText = display.newText ( { text = i, x = 100 + 200 * ( i - 1 ), y = 10, fontSize = 18 } )
	allBoxes[i].iText:setFillColor ( unpack ( color ) )
	allBoxes[i].aboveText = display.newText ( { text = "above: ", x = 100 + 200 * ( i - 1 ), y = 40, fontSize = 18 } )
	allBoxes[i].rightText = display.newText ( { text = "right: ", x = 100 + 200 * ( i - 1 ), y = 70, fontSize = 18 } )
	allBoxes[i].leftText = display.newText ( { text = "left: ", x = 100 + 200 * ( i - 1 ), y = 100, fontSize = 18 } )
	allBoxes[i].inclineText = display.newText ( { text = "incline: ", x = 100 + 200 * ( i - 1 ), y = 130, fontSize = 18 } )
	allBoxes[i].orderText = display.newText ( { text = "order: ", x = 100 + 200 * ( i - 1 ), y = 160, fontSize = 18 } )
	
	allBoxes[i].left = display.newImageRect( allBoxes[i].group, mainName.."_left.png", sideWidth, faceHeight )
	allBoxes[i].left.x = allBoxes[i].face.x - allBoxes[i].face.width/2
	allBoxes[i].left.anchorX = 1
	
	
	allBoxes[i].right = display.newImageRect( allBoxes[i].group, mainName.."_right.png", sideWidth, faceHeight )
	allBoxes[i].right.x = allBoxes[i].face.x + allBoxes[i].face.width/2
	allBoxes[i].right.anchorX = 0

	allBoxes[i].top = display.newImageRect( allBoxes[i].group, mainName.."_top.png", faceWidth, sideHeight )
	allBoxes[i].top.y = allBoxes[i].face.y - allBoxes[i].face.height/2
	allBoxes[i].top.anchorY = 1
	
	allBoxes[i].bottom = display.newImageRect( allBoxes[i].group, mainName.."_bottom.png", faceWidth, sideHeight )
	allBoxes[i].bottom.y = allBoxes[i].face.y + allBoxes[i].face.height/2
	allBoxes[i].bottom.anchorY = 0
	
	
	allBoxes[i].left.alpha, allBoxes[i].right.alpha, allBoxes[i].top.alpha, allBoxes[i].bottom.alpha = 0.5,0.5,0.5,0.5
	--We need 4 key values for box at 0, 90, 180 and 270 degrees rotation.
	--Values between key values will be calculated on run by reDrawBoxes on enterFrame event.

		local LW = allBoxes[i].left.width
		local LH = allBoxes[i].left.height
		local RW = allBoxes[i].right.width
		local RH = allBoxes[i].right.height
		
		local TH = allBoxes[i].top.height
		local BH = allBoxes[i].bottom.height
		
	allBoxes[i].corners = {}
	
	--	at 0 degrees: initial value, siting on bottom
	allBoxes[i].corners[1] = { 
		left = { x1 = 2*LW, y1 = -TH, x2 = 2*LW, y2 = -BH },
		right = { x4 = 0, y4 = -TH, x3 = 0, y3 = -BH },
		top = { x1 = LW, y1 = 0, x4 = RW, y4 = 0 },
		bottom = { x2 = LW, y2 = -2*BH, x3 = RW, y3 = -2*BH }
	}
		--at 90 degr: laying on right side
	allBoxes[i].corners[2] = { 
		left = { x1 = 0, y1 = -TH, x2 = 0, y2 = -BH },
		right = { x4 = -2*RW, y4 = -TH, x3 = -2*RW, y3 = -BH },
		top = { x1 = -LW, y1 = 0, x4 = -RW, y4 = 0 },
		bottom = { x2 = -LW, y2 = -2*BH, x3 = -RW, y3 = -2*BH },
	}
		--at 180 degr: laying on top side
	allBoxes[i].corners[3] = { 
		left = { x1 = 0, y1 = BH, x2 = 0, y2 = BH },
		right = { x4 = -2*RW, y4 = TH, x3 = -2*RW, y3 = BH },
		top = { x1 = -LW, y1 = 2*TH, x4 = -RW, y4 = 2*TH },
		bottom = { x2 = -RW, y2 = 0, x3 = -RW, y3 = 0 },
	}
		--at 270 degr: laying on left side
		allBoxes[i].corners[4] = { 
		left = { x1 = 2*LW, y1 = TH, x2 = 2*LW, y2 = BH },
		right = { x4 = 0, y4 = TH, x3 = 0, y3 = BH },
		top = { x1 = RW, y1 = 2*TH, x4 = LW, y4 = 2*TH },
		bottom = { x2 = LW, y2 = 0, x3 = RW, y3 = 0 },
	}

end
--=========================PHYSICS BODIES
--Add boxes to physics. We will add whole group containing 5 pictures that put together box.
for i = 1, #allBoxes do
	local thisBox = allBoxes[i]
	
	--We need to redo physics body shape so it only covers face of the box.
	local W, H = thisBox.face.width/2, thisBox.face.height/2
	local bodyDesc = { -W,-H, -W,H, W,-H, W,H }
	
	-- physics.addBody ( thisBox.group, "dynamic", { bounce = 0.05, density = 3, friction = 1, shape = bodyDesc } )
	-- thisBox.group.isBullet = true
	thisBox.group.isAwake = true
	thisBox.group:addEventListener ( "touch", touchBox )
end

--========================RUNTIME enterFrame LISTENER
local function reDrawBoxes ()
--this is function that recalculates visual of all boxes in real time and applies it using rearangeCorners and rearangeDrawOrder functions.
	for i = 1, #allBoxes do
		local thisBox = allBoxes[i]
			thisBox.aboveText.text = "above: "
			thisBox.rightText.text = "right: "
			thisBox.leftText.text = "left: "
		
		if thisBox.group.isAwake then
			local rotation = allBoxes[i].group.rotation
			local angle = rotation % 360 --value that represents the same visual rotation but in interval 0 to 360
			
			
			local startTable, endTable --key values will be stored here
			local interval --we will calculate values for each interval of 90...this will contain angle value recalculated to this interval
			
			if angle < 90 then
				startTable = thisBox.corners[1]
				endTable = thisBox.corners[2]
				interval = angle
				
				thisBox.top:toFront()
				thisBox.bottom:toBack()
			elseif angle >=90 and angle < 180 then
				startTable = thisBox.corners[2]
				endTable = thisBox.corners[3]
				interval = angle - 90
				
				thisBox.left:toFront()
				thisBox.right:toBack()
			elseif angle >= 180 and angle < 270 then
				startTable = thisBox.corners[3]
				endTable = thisBox.corners[4]
				interval = angle - 180
				
				thisBox.bottom:toFront()
				thisBox.top:toBack()
			elseif angle >=270 and angle < 360 then
				startTable = thisBox.corners[4]
				endTable = thisBox.corners[1]
				interval = angle - 270
				
				thisBox.right:toFront()
				thisBox.left:toBack()
			end
			thisBox.face:toFront()
			
			local actualTable = {} --actual calculated values for given rotation
			
			for img, valuesTable in pairs ( startTable ) do
				
				actualTable[img] = {}

				for prop, value in pairs ( valuesTable ) do 

					local startTableValue = startTable[img][prop] --value in startTable table
					local endTableValue = endTable[img][prop]
					
					local actualTableValue = ( endTableValue - startTableValue )*( interval / 90 ) + startTableValue
					
					--put calculated actualTableValue to table
					actualTable[img][prop] = actualTableValue
				end
			end
			--apply calculated values to thisBox corners
			rearangeCorners ( thisBox, actualTable )
		end
	end --end of loop for allBoxes
	
	--redo order in which are objects on screen. We need faces to be always in front of sides.
	rearangeDrawOrder ()
end

Runtime:addEventListener( "enterFrame", reDrawBoxes )

local function keyPress ( event )
	if event.phase == "up" then
		if event.keyName == "space" then
			printRuntime = false
		elseif event.keyName == "c" then
			printRuntime = true
		elseif event.keyName == "i" then
			P = true
		elseif event.keyName == "q" then
			allBoxes[1].group:rotate ( -5 )
		elseif event.keyName == "e" then
			allBoxes[1].group:rotate ( 5 )
		elseif event.keyName == "r" then
			allBoxes[2].group:rotate ( -5 )
		elseif event.keyName == "z" then
			allBoxes[2].group:rotate ( 5 )
		elseif event.keyName == "w" then
			allBoxes[1].group:toFront ()
		elseif event.keyName == "t" then
			allBoxes[2].group:toFront ()
		end
	end
	return true
end

Runtime:addEventListener ( "key", keyPress )